#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "lib.h"
#include "line.h"
#include "variable.h"
#include "label.h"
#include "position.h"
#include "readline.h"
#include "nll.h"

static void sigint_handler(int arg)
{
  nll_finish();
  return;
}

static void signal_init(void)
{
  struct sigaction sa;

  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = sigint_handler;

  sigaction(SIGINT, &sa, NULL);
}

static void signal_reset(void)
{
  struct sigaction sa;

  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = SIG_DFL;

  sigaction(SIGINT, &sa, NULL);
}

static void init(void)
{
  variable_init();
  position_init();
  label_init();
  line_init();
  signal_init();
}

static void done(void)
{
  signal_reset();
}

static void truncate_line(char *buffer)
{
  int len;
  if (buffer) {
    len = strlen(buffer);
    while (len > 0) {
      if (!isspace(buffer[len - 1]))
	break;
      len--;
    }
    buffer[len] = '\0';
  }
}

static int readfile(const char *filename)
{
  FILE *fp;
  char buffer[LINE_MAXLEN + 1];
  int r, n = 0;
  line_t line;

  fp = fopen(filename, "r");
  if (!fp)
    NLL_ERREXIT(NLL_ERRCODE_FILE_NOT_FOUND, filename);

  while (fgets(buffer + 1, sizeof(buffer) - 1, fp) != NULL) {
    buffer[0] = ' ';
    truncate_line(buffer);
    if ((r = line_set(++n, -1, buffer, &line)) < 0)
      return r;
    if (line)
      nll_check(line, 0);
  }

  fclose(fp);

  return 0;
}

int main(int argc, char *argv[])
{
  char *filename = NULL, *buffer, *top, *p;
  int finished = 0, r, number, number2;
  line_t line;

  argc--;
  argv++;

  if (argc > 0) {
    filename = argv[0];
  }

  init();

  if (filename) {
    if ((r = readfile(filename)) < 0)
      NLL_ERREXIT(r, filename);
    if (nll_run(line_head(), 0) < 0)
      finished = 1;
  }

  while (!finished) {
    buffer = nll_readline("nll> ");
    if (!buffer) {
      printf("\n");
      continue;
    }

    truncate_line(buffer);

    p = buffer;
    while (isspace(*p))
      p++;

    number = strtoul(p, &top, 10);

    if (top == p) { /* direct mode */
      if ((r = line_set(-1, -1, top, &line)) < 0)
	NLL_ERREXIT(r, top);
      if (line) {
	nll_check(line, 0);
	nll_add_history(line->buffer);
	if (nll_run(line, 0) < 0)
	  finished = 1;
	if ((r = line_free(line)) < 0)
	  NLL_ERREXIT(r, top);
      }
    } else { /* program mode */
      nll_add_history(p);
      p = top;
      while (isspace(*p))
	p++;
      number2 = isdigit(*p) ? strtoul(p, &top, 10) : -1;
      if ((r = line_set(number, number2, top, &line)) < 0)
	NLL_ERREXIT(r, top);
      if (line)
	nll_check(line, 0);
    }

    free(buffer);
  }

  done();

  return 0;
}
