#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "lib.h"
#include "line.h"
#include "variable.h"
#include "value.h"
#include "word.h"
#include "command.h"
#include "label.h"
#include "position.h"
#include "readline.h"
#include "nll.h"

static int stepping = 0;

static line_t break_line = NULL;
static int break_column = 0;

static line_t continue_line = NULL;
static int continue_column = 0;

static int clear_break(line_t line)
{
  if (!line || (break_line == line)) {
    break_line = NULL;
    break_column = 0;
  }
  return 0;
}

static int clear_continue(line_t line)
{
  if (!line || (continue_line == line)) {
    continue_line = NULL;
    continue_column = 0;
  }
  return 0;
}

int nll_clear_line(line_t line)
{
  clear_break(line);
  clear_continue(line);
  return 0;
}

static int get_continue(line_t *linep, int *columnp)
{
  if (continue_line == NULL)
    return NLL_ERRCODE_VALUE_NO_CONTINUE_POINT;
  *linep   = continue_line;
  *columnp = continue_column;
  return 0;
}

static int set_continue(line_t line, int column)
{
  continue_line   = line;
  continue_column = column;
  return 0;
}

static int proc_let(const char *variable, const char *value)
{
  int r, integer = 0;
  if (value[0]) {
    if ((r = value_get(value, &integer)) < 0)
      return r;
  }
  if ((r = variable_set_integer(variable, integer)) < 0)
    return r;
  return 0;
}

static int proc_print(const char *value)
{
  int r, integer;
  if (value[0]) {
    if ((r = value_get(value, &integer)) < 0)
      return r;
    printf("%d", integer);
    fflush(stdout);
  }
  return 0;
}

static int proc_go(line_t *linep, int *columnp, const char *variable, const char *value)
{
  int r, integer = 1;
  label_t label;

  if (value[0]) {
    if ((r = value_get(value, &integer)) < 0)
      return r;
  }

  if (integer) {
    if ((r = label_get(variable, &label)) < 0)
      return r;
    *linep   = label->line;
    *columnp = label->column;
  }

  return integer ? 1 : 0;
}

static int proc_for(line_t line, int column, const char *variable, const char *value0, const char *value1, const char *value2, position_param_t param)
{
  int r, integer0 = 10, integer1 = 0, integer2 = 1;

  if (value0[0]) {
    if ((r = value_get(value0, &integer0)) < 0)
      return r;
  }

  if (value1[0]) {
    if ((r = value_get(value1, &integer1)) < 0)
      return r;
  }

  if (value2[0]) {
    if ((r = value_get(value2, &integer2)) < 0)
      return r;
  }

  if ((r = variable_set_integer(variable, integer1)) < 0)
    return r;

  strcpy(param->_for.variable, variable);
  param->_for.limit = integer1 + integer2 * integer0;
  param->_for.step  = integer2;

  if ((r = position_push(POSITION_TYPE_FOR, line, column, param)) < 0)
    return r;

  return 0;
}

static int proc_next(line_t *linep, int *columnp, const char *variable, position_param_t param)
{
  int r, integer;
  line_t line;
  int column;

  if ((r = position_pop(POSITION_TYPE_FOR, &line, &column, param)) < 0)
    return r;

  if (variable[0]) {
    if (strcmp(param->_for.variable, variable))
      return NLL_ERRCODE_POSITION_DIFFERENT_VARIABLE;
  }

  if ((r = variable_get_integer(param->_for.variable, &integer)) < 0)
    return r;

  integer += param->_for.step;

  if ((r = variable_set_integer(param->_for.variable, integer)) < 0)
    return r;

  if (integer != param->_for.limit) {
    if ((r = position_push(POSITION_TYPE_FOR, line, column, param)) < 0)
      return r;
    *linep   = line;
    *columnp = column;
    return 1;
  }

  return 0;
}

static int proc_run(line_t *linep, int *columnp, const char *value)
{
  int r, integer = 1;
  line_t line = NULL;

  if (value[0]) {
    if ((r = value_get(value, &integer)) < 0)
      return r;
    if ((r = line_get(integer, &line)) < 0)
      return r;
  }

  if (!line)
    line = line_head();

  *linep   = line;
  *columnp = 0;

  return 0;
}

static int proc_break(const char *value0, const char *value1)
{
  int r, integer0 = 0, integer1 = 0;
  line_t line = NULL;

  if (value0[0]) {
    if ((r = value_get(value0, &integer0)) < 0)
      return r;
    if ((r = line_get(integer0, &line)) < 0)
      return r;
  }

  if (value1[0]) {
    if ((r = value_get(value1, &integer1)) < 0)
      return r;
  }

  if (!line) {
    clear_break(NULL);
  } else {
    break_line = line;
    break_column = integer1;
  }

  return 0;
}

static int proc_step(const char *value)
{
  int r, integer = 1;

  if (value[0]) {
    if ((r = value_get(value, &integer)) < 0)
      return r;
  }

  stepping = integer;

  return 0;
}

static int proc_list(const char *value0, const char *value1)
{
  int r, integer0 = -1, integer1 = -1;

  if (value0[0]) {
    if ((r = value_get(value0, &integer0)) < 0)
      return r;
  }

  if (value1[0]) {
    if ((r = value_get(value1, &integer1)) < 0)
      return r;
  } else {
    integer1 = integer0;
  }

  line_print(integer0, integer1);
  fflush(stdout);

  return 0;
}

static int proc_edit(const char *value)
{
  int r, integer = 0;
  line_t line;

  if (value[0]) {
    if ((r = value_get(value, &integer)) < 0)
      return r;
  }

  if ((r = line_get(integer, &line)) < 0)
    return r;

  if (line) {
    nll_edit_line(integer, line);
  }

  return 0;
}

static int proc_wait(const char *value)
{
  int r, integer = 100;
  if (value[0]) {
    if ((r = value_get(value, &integer)) < 0)
      return r;
  }
  usleep(integer * 10000);
  return 0;
}

static int _nll_run(line_t line, int column, line_t *linep, int *columnp, int execute)
{
  char *p, *tail, *buffer, *pc;
  command_type_t type;
#define ARG_LEN 256
  char arg0[ARG_LEN], arg1[ARG_LEN], arg2[ARG_LEN], arg3[ARG_LEN];
  int r, count = 0;
  union position_param param;

  buffer = line->buffer;
  p = buffer + column;
  tail = p + strlen(p);

#define ERRBREAK(r, param) \
	do { NLL_ERRPRINT(r, param); goto brk; } while(0)

  while (p < tail) {
    pc = p;

    if ((r = word_get_word(&p, arg0, ARG_LEN)) < 0)
      ERRBREAK(r, buffer);
    if (arg0[0] == '\0') {
      p++;
      continue;
    }

    if (execute) {
      if (nll_is_finished()) {
	fprintf(stderr, "\nBreak at: %s\n", pc);
	goto brk;
      }

      if (break_line && (line_is_exist(line) > 0)) {
	if ((line == break_line) && (pc - buffer >= break_column)) {
	  clear_break(line);
	  fprintf(stderr, "\nBreakpoint at: %s\n", pc);
	  goto brk;
	}
      }

      if (stepping && (line_is_exist(line) > 0)) {
	stepping--;
	if (!stepping) {
	  fprintf(stderr, "\nStep break at: %s\n", pc);
	  goto brk;
	}
      }
    }

    count++;

    if (*p == ':') {
      p++;
      if ((r = label_set(arg0, line, p - buffer)) < 0)
	ERRBREAK(r, buffer);
      continue;
    }

    type = search_command(arg0);

    switch (type) {
    case COMMAND_COMMENT:
      goto nextline;

    case COMMAND_END:
      if (!execute)
	break;
      goto end;

    case COMMAND_EXIT:
      if (!execute)
	break;
      goto ext;

    case COMMAND_LET:
      if ((r = word_get_word(&p, arg0, ARG_LEN)) < 0) ERRBREAK(r, buffer);
      if ((r = word_get_value(&p, arg1, ARG_LEN)) < 0) ERRBREAK(r, buffer);
      if (!execute)
	break;
      if ((r = proc_let(arg0, arg1)) < 0) ERRBREAK(r, buffer);
      break;

    case COMMAND_PRINT:
      if ((r = word_get_value(&p, arg0, ARG_LEN)) < 0) ERRBREAK(r, buffer);
      if (!execute)
	break;
      if ((r = proc_print(arg0)) < 0) ERRBREAK(r, buffer);
      break;

    case COMMAND_GOTO:
    case COMMAND_GOSUB:
      if ((r = word_get_word(&p, arg0, ARG_LEN)) < 0) ERRBREAK(r, buffer);
      if ((r = word_get_value(&p, arg1, ARG_LEN)) < 0) ERRBREAK(r, buffer);
      if (!execute)
	break;
      if ((r = proc_go(linep, columnp, arg0, arg1)) < 0) ERRBREAK(r, buffer);
      if (!r)
	break;
      if (type == COMMAND_GOSUB) {
	if ((r = position_push(POSITION_TYPE_GOSUB, line, p + 1 - buffer, NULL)) < 0) ERRBREAK(r, buffer);
      }
      goto next;

    case COMMAND_RETURN:
      if (!execute)
	break;
      if ((r = position_pop(POSITION_TYPE_GOSUB, linep, columnp, NULL)) < 0) ERRBREAK(r, buffer);
      goto next;

    case COMMAND_FOR:
      if ((r = word_get_word(&p, arg0, ARG_LEN)) < 0) ERRBREAK(r, buffer);
      if ((r = word_get_value(&p, arg1, ARG_LEN)) < 0) ERRBREAK(r, buffer);
      if ((r = word_get_value(&p, arg2, ARG_LEN)) < 0) ERRBREAK(r, buffer);
      if ((r = word_get_value(&p, arg3, ARG_LEN)) < 0) ERRBREAK(r, buffer);
      if (!execute)
	break;
      if ((r = proc_for(line, p + 1 - buffer, arg0, arg1, arg2, arg3, &param)) < 0) ERRBREAK(r, buffer);
      break;

    case COMMAND_NEXT:
      if ((r = word_get_word(&p, arg0, ARG_LEN)) < 0) ERRBREAK(r, buffer);
      if (!execute)
	break;
      if ((r = proc_next(linep, columnp, arg0, &param)) < 0) ERRBREAK(r, buffer);
      if (!r)
	break;
      goto next;

    case COMMAND_RUN:
      if ((r = word_get_value(&p, arg0, ARG_LEN)) < 0) ERRBREAK(r, buffer);
      if (!execute)
	break;
      if ((r = proc_run(linep, columnp, arg0)) < 0) ERRBREAK(r, buffer);
      variable_init();
      position_init();
      stepping = 0;
      goto next;

    case COMMAND_STOP:
      if (!execute)
	break;
      if (*p)
	p++;
      pc = p;
      fprintf(stderr, "\nStop at: %s\n", pc);
      goto brk;

    case COMMAND_BREAK:
      if ((r = word_get_value(&p, arg0, ARG_LEN)) < 0) ERRBREAK(r, buffer);
      if ((r = word_get_value(&p, arg1, ARG_LEN)) < 0) ERRBREAK(r, buffer);
      if (!execute)
	break;
      if ((r = proc_break(arg0, arg1)) < 0) ERRBREAK(r, buffer);
      break;

    case COMMAND_STEP:
      if ((r = word_get_value(&p, arg0, ARG_LEN)) < 0) ERRBREAK(r, buffer);
      if (!execute)
	break;
      if ((r = proc_step(arg0)) < 0) ERRBREAK(r, buffer);
      if (stepping) stepping++;
      /* fall through */

    case COMMAND_CONTINUE:
      if (!execute)
	break;
      if ((r = get_continue(linep, columnp)) < 0) ERRBREAK(r, buffer);
      clear_continue(NULL);
      goto next;

    case COMMAND_NEW:
      if (!execute)
	break;
      position_init();
      label_init();
      line_init();
      break;

    case COMMAND_LIST:
      if ((r = word_get_value(&p, arg0, ARG_LEN)) < 0) ERRBREAK(r, buffer);
      if ((r = word_get_value(&p, arg1, ARG_LEN)) < 0) ERRBREAK(r, buffer);
      if (!execute)
	break;
      if ((r = proc_list(arg0, arg1)) < 0) ERRBREAK(r, buffer);
      break;

    case COMMAND_EDIT:
      if ((r = word_get_value(&p, arg0, ARG_LEN)) < 0) ERRBREAK(r, buffer);
      if (!execute)
	break;
      if ((r = proc_edit(arg0)) < 0) ERRBREAK(r, buffer);
      break;

    case COMMAND_WAIT:
      if ((r = word_get_value(&p, arg0, ARG_LEN)) < 0) ERRBREAK(r, buffer);
      if (!execute)
	break;
      if ((r = proc_wait(arg0)) < 0) ERRBREAK(r, buffer);
      break;

    case COMMAND_UNKNOWN:
    default:
      break;
    }

    if (*p)
      p++;
  }

nextline:
  *linep = line->next;
  *columnp = 0;

next:
  return count;

ext:
  *linep = NULL;
  return -1;

end:
  *linep = NULL;
  return 0;

brk:
  if ((r = set_continue(line, pc - buffer)) < 0) ERRBREAK(r, buffer);
  *linep = NULL;
  return count;
}

int nll_check(line_t line, int column)
{
  return _nll_run(line, column, &line, &column, 0);
}

int nll_run(line_t line, int column)
{
  int r = 0;

  nll_finished_clear();

  while (line) {
    r = _nll_run(line, column, &line, &column, 1);
  }

  return r;
}
