#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "lib.h"
#include "variable.h"
#include "value.h"

typedef enum operator_type {
  OPERATOR_TYPE_INC   ,
  OPERATOR_TYPE_DEC   ,
  OPERATOR_TYPE_NONE  ,
  OPERATOR_TYPE_LSHIFT,
  OPERATOR_TYPE_RSHIFT,
  OPERATOR_TYPE_LE    ,
  OPERATOR_TYPE_GE    ,
  OPERATOR_TYPE_EQ    ,
  OPERATOR_TYPE_NE    ,
  OPERATOR_TYPE_LOGAND,
  OPERATOR_TYPE_LOGOR ,
  OPERATOR_TYPE_NOT   ,
  OPERATOR_TYPE_INV   ,
  OPERATOR_TYPE_PLUS  ,
  OPERATOR_TYPE_MINUS ,
  OPERATOR_TYPE_MUL   ,
  OPERATOR_TYPE_DIV   ,
  OPERATOR_TYPE_MOD   ,
  OPERATOR_TYPE_ADD   ,
  OPERATOR_TYPE_SUB   ,
  OPERATOR_TYPE_LT    ,
  OPERATOR_TYPE_GT    ,
  OPERATOR_TYPE_XOR   ,
  OPERATOR_TYPE_AND   ,
  OPERATOR_TYPE_OR    ,
  OPERATOR_TYPE_SUBST ,
} operator_type_t;

typedef enum operator_subtype {
  OPERATOR_SUBTYPE_NONE,
  OPERATOR_SUBTYPE_MONADIC,
  OPERATOR_SUBTYPE_MONADIC_BINARY,
  OPERATOR_SUBTYPE_BINARY,
} operator_subtype_t;

typedef struct operator {
  char *word;
  int priority;
  operator_type_t type;
  operator_subtype_t subtype;
} *operator_t;

struct operator operators[] = {
  { "++" , 1, OPERATOR_TYPE_INC   , OPERATOR_SUBTYPE_MONADIC },
  { "--" , 1, OPERATOR_TYPE_DEC   , OPERATOR_SUBTYPE_MONADIC },
  { "<<",  4, OPERATOR_TYPE_LSHIFT, OPERATOR_SUBTYPE_BINARY  },
  { ">>",  4, OPERATOR_TYPE_RSHIFT, OPERATOR_SUBTYPE_BINARY  },
  { "<=",  5, OPERATOR_TYPE_LE    , OPERATOR_SUBTYPE_BINARY  },
  { ">=",  5, OPERATOR_TYPE_GE    , OPERATOR_SUBTYPE_BINARY  },
  { "==",  6, OPERATOR_TYPE_EQ    , OPERATOR_SUBTYPE_BINARY  },
  { "!=",  6, OPERATOR_TYPE_NE    , OPERATOR_SUBTYPE_BINARY  },
  { "<>",  6, OPERATOR_TYPE_NE    , OPERATOR_SUBTYPE_BINARY  },
  { "&&", 10, OPERATOR_TYPE_LOGAND, OPERATOR_SUBTYPE_BINARY  },
  { "||", 11, OPERATOR_TYPE_LOGOR , OPERATOR_SUBTYPE_BINARY  },
  { ":=", 12, OPERATOR_TYPE_SUBST , OPERATOR_SUBTYPE_BINARY  },
  { "<-", 12, OPERATOR_TYPE_SUBST , OPERATOR_SUBTYPE_BINARY  },
  { "!" ,  1, OPERATOR_TYPE_NOT   , OPERATOR_SUBTYPE_MONADIC },
  { "~" ,  1, OPERATOR_TYPE_INV   , OPERATOR_SUBTYPE_MONADIC },
  { "+" ,  1, OPERATOR_TYPE_PLUS  , OPERATOR_SUBTYPE_MONADIC_BINARY },
  { "-" ,  1, OPERATOR_TYPE_MINUS , OPERATOR_SUBTYPE_MONADIC_BINARY },
  { "*" ,  2, OPERATOR_TYPE_MUL   , OPERATOR_SUBTYPE_BINARY  },
  { "/" ,  2, OPERATOR_TYPE_DIV   , OPERATOR_SUBTYPE_BINARY  },
  { "%" ,  2, OPERATOR_TYPE_MOD   , OPERATOR_SUBTYPE_BINARY  },
  { "+" ,  3, OPERATOR_TYPE_ADD   , OPERATOR_SUBTYPE_BINARY  },
  { "-" ,  3, OPERATOR_TYPE_SUB   , OPERATOR_SUBTYPE_BINARY  },
  { "<" ,  5, OPERATOR_TYPE_LT    , OPERATOR_SUBTYPE_BINARY  },
  { ">" ,  5, OPERATOR_TYPE_GT    , OPERATOR_SUBTYPE_BINARY  },
  { "^" ,  8, OPERATOR_TYPE_XOR   , OPERATOR_SUBTYPE_BINARY  },
  { "&" ,  7, OPERATOR_TYPE_AND   , OPERATOR_SUBTYPE_BINARY  },
  { "|" ,  9, OPERATOR_TYPE_OR    , OPERATOR_SUBTYPE_BINARY  },
  { "=" , 12, OPERATOR_TYPE_SUBST , OPERATOR_SUBTYPE_BINARY  },
  { NULL,  0, OPERATOR_TYPE_NONE  , OPERATOR_SUBTYPE_NONE    },
};

#define OPERATOR_PRIORITY_MIN  1
#define OPERATOR_PRIORITY_MAX 12
#define OPERATOR_PRIORITY_NUM \
	(OPERATOR_PRIORITY_MAX - OPERATOR_PRIORITY_MIN + 1)

#define ELEMENT_MAXNUM 64

typedef enum {
  ELEMENT_TYPE_NONE,
  ELEMENT_TYPE_INTEGER,
  ELEMENT_TYPE_STRING,
  ELEMENT_TYPE_VARIABLE,
  ELEMENT_TYPE_OPERATOR,
} element_type_t;

typedef union element_param {
  int integer;
  struct operator *operator;
} *element_param_t;

typedef struct element {
  element_type_t type;
  union element_param param;
} *element_t;

static struct element elements[ELEMENT_MAXNUM];
static int element_num;

static operator_t operator_get(operator_type_t type)
{
  operator_t operator;
  for (operator = operators; operator->word; operator++) {
    if (operator->type == type)
      return operator;
  }
  return NULL;
}

static int element_init(void)
{
  element_num = 0;
  return 0;
}

static int element_push(element_type_t type, element_param_t param)
{
  element_t e;
  if (element_num >= ELEMENT_MAXNUM)
    return NLL_ERRCODE_VALUE_STACK_OVER;
  e = &elements[element_num++];
  e->type = type;
  memcpy(&e->param, param, sizeof(e->param));
  return 0;
}

#if 0
static int element_pop(int num)
{
  if (element_num < num)
    return NLL_ERRCODE_VALUE_LESS_PARAMETER;
  element_num -= num;
  return 0;
}
#endif

static int element_cut(int index, int num)
{
  if (index + num > element_num)
    return NLL_ERRCODE_VALUE_LESS_PARAMETER;
  for (; index < element_num - num; index++)
    memcpy(&elements[index], &elements[index + num], sizeof(elements[0]));
  element_num -= num;
  return 0;
}

static int element_get_integer(int index, int *integerp)
{
  if (index >= element_num)
    return NLL_ERRCODE_VALUE_LESS_PARAMETER;
  if (elements[index].type != ELEMENT_TYPE_INTEGER)
    return NLL_ERRCODE_VALUE_INVALID_VALUE;
  *integerp = elements[index].param.integer;
  return 0;
}

static int element_set_integer(int index, int integer)
{
  if (index >= element_num)
    return NLL_ERRCODE_VALUE_LESS_PARAMETER;
  elements[index].type = ELEMENT_TYPE_INTEGER;
  elements[index].param.integer = integer;
  return 0;
}

static int a2val(char c)
{
  if (isdigit(c))
    return c - '0';
  else if (isalpha(c))
    return toupper(c) - 'A' + 10;
  return -1;
}

static int value_read(const char *value, const char **endp, int terminator)
{
  int r, start_index, index, priority, integer, i, v, x, y, scale;
  const char *p;
  char word[16];
  element_t e, el, er;
  operator_type_t opetype;
  union element_param param;
  operator_t operator;

  p = value;

  start_index = element_num;

  while (1) {
    if (*p == '\0')
      break;
    else if (isspace(*p)) {
      p++;
      continue;
    } else if (*p == terminator) {
      p++;
      break;
    } else if (isdigit(*p)) { /* integer */
      integer = 0;
      scale = 10;
      if (*p == '0') {
	scale = 8;
	p++;
	if ((*p == 'x') || (*p == 'X')) {
	  scale = 16;
	  p++;
	} else if ((*p == 'b') || (*p == 'B')) {
	  scale = 2;
	  p++;
	}
      }
      while (1) {
	v = a2val(*p);
	if (v < 0)
	  break;
	if (v >= scale)
	  break;
	integer = integer * scale + v;
	p++;
      }
      param.integer = integer;
      if ((r = element_push(ELEMENT_TYPE_INTEGER, &param)) < 0)
	return r;
    } else if (isalpha(*p)) { /* variable */
      for (i = 0; isalnum(*p); i++, p++)
	word[i] = *p;
      word[i] = '\0';
      if ((r = variable_get_integer(word, &integer)) < 0) return r;
      param.integer = integer;
      if ((r = element_push(ELEMENT_TYPE_INTEGER, &param)) < 0)
	return r;
    } else if (*p == '(') {
      p++;
      if ((r = value_read(p, &p, ')')) < 0) return r;
    } else if (*p == '[') {
      p++;
      if ((r = value_read(p, &p, ']')) < 0) return r;
    } else { /* operator */
      for (operator = operators; operators->word; operator++) {
	if (!strncmp(operator->word, p, strlen(operator->word)))
	  break;
      }
      if (!operators->word)
	return NLL_ERRCODE_VALUE_INVALID_OPERATOR;
      param.operator = operator;
      if ((r = element_push(ELEMENT_TYPE_OPERATOR, &param)) < 0)
	return r;
      p += strlen(operator->word);
    }
  }

  for (i = 0; i < OPERATOR_PRIORITY_NUM; i++) {
    priority = OPERATOR_PRIORITY_MIN + i;

    for (index = start_index; index < element_num; index++) {
      e = &elements[index];
      el = (index     > start_index) ? &elements[index - 1] : NULL;
      er = (index + 1 < element_num) ? &elements[index + 1] : NULL;

      if (e->type != ELEMENT_TYPE_OPERATOR)
	continue;
      if (e->param.operator->priority != priority)
	continue;

      opetype = e->param.operator->type;

      switch (e->param.operator->subtype) {
      case OPERATOR_SUBTYPE_MONADIC_BINARY:
	if (el && (el->type != ELEMENT_TYPE_OPERATOR)) {
	  switch (opetype) {
	  case OPERATOR_TYPE_PLUS:
	    e->param.operator = operator_get(OPERATOR_TYPE_ADD); break;
	  case OPERATOR_TYPE_MINUS:
	    e->param.operator = operator_get(OPERATOR_TYPE_SUB); break;
	  default:
	    return NLL_ERRCODE_VALUE_UNSUPPORTED_OPERATOR;
	  }
	  break;
	}
	/* Fall through */

      case OPERATOR_SUBTYPE_MONADIC:
	if (!er)
	  return NLL_ERRCODE_VALUE_LESS_PARAMETER;
	if ((r = element_get_integer(index + 1, &integer)) < 0) return r;
	switch (opetype) {
	case OPERATOR_TYPE_INC:   integer++;          break;
	case OPERATOR_TYPE_DEC:   integer--;          break;
	case OPERATOR_TYPE_PLUS:                      break;
	case OPERATOR_TYPE_MINUS: integer = -integer; break;
	case OPERATOR_TYPE_NOT:   integer = !integer; break;
	case OPERATOR_TYPE_INV:   integer = ~integer; break;
	default:
	  return NLL_ERRCODE_VALUE_UNSUPPORTED_OPERATOR;
	}
	if ((r = element_set_integer(index, integer)) < 0) return r;
	if ((r = element_cut(index + 1, 1)) < 0) return r;
	break;

      case OPERATOR_SUBTYPE_BINARY:
	if (!el || !er)
	  return NLL_ERRCODE_VALUE_LESS_PARAMETER;
	if ((r = element_get_integer(index - 1, &x)) < 0) return r;
	if ((r = element_get_integer(index + 1, &y)) < 0) return r;
	switch (opetype) {
        case OPERATOR_TYPE_DIV:
        case OPERATOR_TYPE_MOD:
	  if (y == 0)
	    return NLL_ERRCODE_VALUE_DIV_ZERO;
	  break;
	default:
	  break;
	}
	switch (opetype) {
	case OPERATOR_TYPE_LSHIFT: x = x << y; break;
	case OPERATOR_TYPE_RSHIFT: x = x >> y; break;
	case OPERATOR_TYPE_LE:     x = x <= y; break;
	case OPERATOR_TYPE_GE:     x = x >= y; break;
	case OPERATOR_TYPE_EQ:     x = x == y; break;
	case OPERATOR_TYPE_NE:     x = x != y; break;
	case OPERATOR_TYPE_LOGAND: x = x && y; break;
	case OPERATOR_TYPE_LOGOR:  x = x || y; break;
	case OPERATOR_TYPE_MUL:    x = x *  y; break;
	case OPERATOR_TYPE_DIV:    x = x /  y; break;
	case OPERATOR_TYPE_MOD:    x = x %  y; break;
	case OPERATOR_TYPE_ADD:    x = x +  y; break;
	case OPERATOR_TYPE_SUB:    x = x -  y; break;
	case OPERATOR_TYPE_LT:     x = x <  y; break;
	case OPERATOR_TYPE_GT:     x = x >  y; break;
	case OPERATOR_TYPE_XOR:    x = x ^  y; break;
	case OPERATOR_TYPE_AND:    x = x &  y; break;
	case OPERATOR_TYPE_OR:     x = x |  y; break;

	case OPERATOR_TYPE_SUBST:
	default:
	  return NLL_ERRCODE_VALUE_UNSUPPORTED_OPERATOR;
	}
	if ((r = element_set_integer(index - 1, x)) < 0) return r;
	if ((r = element_cut(index, 2)) < 0) return r;
	index--;
	break;

      case OPERATOR_SUBTYPE_NONE:
      default:
	return NLL_ERRCODE_VALUE_UNSUPPORTED_OPERATOR;
      }
    }
  }

  *endp = p;

  return 0;
}

int value_get(const char *value, int *integerp)
{
  int r;
  element_init();
  if ((r = value_read(value, &value, ',')) < 0)
    return r;
  if (element_num == 0)
    return NLL_ERRCODE_VALUE_LESS_PARAMETER;
  *integerp = elements[0].param.integer;
  return 0;
}
